//
//  $Id: SimpleList.m,v 1.2 2008/04/24 08:09:37 rh Exp $
//  A simple List class implementation
//
//  Copyright 2006-2008 Rene Hexel. All rights reserved.
//

#import "SimpleList.h"
#include <stdarg.h>

/*
 * private inner class for a list node
 */
@interface _ListNode: NSObject
{
	_ListNode *prev;	// previous Node
	_ListNode *next;	// next Node
	id object;		// contained object
}

- prev;				// return and next previous nodes
- next;
- object;			// return object (content)
- setPrev: (_ListNode *) newPrev; // set previous and next nodes
- setNext: (_ListNode *) newNext;
- setObject: (id) newObj;	// set new content

- (void) dealloc;		// return all memory

@end

/*
 * implementation of the node class
 */
@implementation _ListNode

- prev
{
	return prev;
}


- next
{
	return next;
}


- object
{
	return object;
}


- setPrev: (_ListNode *) newPrev
{
	id old = prev;
	prev = newPrev;

	return old;
}

- setNext: (_ListNode *) newNext
{
	id old = next;
	next = newNext;
	
	return old;
}


- setObject: (id) newObj
{
	[object release];
	object = [newObj retain];
	return self;
}


- (void) dealloc
{
	[object release];
	[super dealloc];
}


@end


/*
 * the actual list class
 */
@implementation SimpleList

/*
 * initialise List with objects
 */
- initWithObjects: (id) obj, ...
{
	va_list ap;				// variable argument pointer
	
	if (!(self = [self init]))		// create empty list
		return nil;

	va_start(ap, obj);
	while (obj)				// for each passed argument
	{
		[self addObject: obj];
		obj = va_arg(ap, id);		// next argument
	}
	va_end(ap);
	
	return self;
}


/*
 * add an object to the tail of the list
 */
- addObject: (id) obj
{
	_ListNode *node = [_ListNode new];	// allocate new node
	_ListNode *prev = [self tail];
	[node setPrev: prev];			// set previous node
	[node setObject: obj];			// set the object
	[prev setNext: node];			// successor of previous node
	[self setTail: node];			// also the new tail of the list
	if (!head) [self setHead: node];	// head if list was empty

	return self;
}


- (ListEnumerator *) objectEnumerator
{
	return [[[ListEnumerator alloc] initWithList: self] autorelease];
}


- (void) dealloc
{
	NSAutoreleasePool *pool = [NSAutoreleasePool new];
	ListEnumerator *e = [self objectEnumerator];
	while ([e nextObject]) [e removeObject];
	[pool release];
	[super dealloc];
}


- head
{
	return head;
}


- tail
{
	return tail;
}


- setHead: (id) newHead
{
	id oldHead = head;
	head = newHead;
	return oldHead;
}


- setTail: (id) newTail
{
	id oldTail = tail;
	tail = newTail;
	return oldTail;
}


@end

/*
 * private methods
 */
@interface SimpleList (Private)
- (_ListNode *) removeNode: (_ListNode *) node;
@end

@implementation SimpleList (Private)

- (_ListNode *) removeNode: (_ListNode *) node
{
	_ListNode *pred = [node prev];	// the predecessor and
	_ListNode *succ = [node next];	// the successor
	[pred setNext: succ];		// replace with successor in predecessor
	[succ setPrev: pred];		// replace with predecessor in successor
	if (node == head) head = succ;	// adjust head and tail if necessary
	if (node == tail) tail = pred;
	
	return [node autorelease];	// release the node
}

@end


/*
 * An object enumerator for the above list class
 */

@implementation ListEnumerator: NSEnumerator

- advance
{
	currentNode = nextNode;			// set current node
	nextNode = [currentNode next];		// advance in the list
	
	return currentNode ? self : nil;	// nil if at end
}


- nextObject
{
	[self advance];

	return [currentNode object];		// return contained object
}


- initWithList: (SimpleList *) newList
{
	if (self = [[self init] setList: newList])
		nextNode = [list head];

	return self;
}


- setList: (SimpleList *) newList
{
	if (newList != list)
	{
		[list release];
		list = [newList retain];
	}

	return self;
}


- removeObject
{
	return [[list removeNode: currentNode] object];
}


- (void) dealloc
{
	[list release];
	[super dealloc];
}
@end
